import os
import re
import subprocess

from .svc import svc


# --------------------
## Operating System Abstraction Layer; provides functions to make cross-platform behavior similar
class Osal:
    # --------------------
    ## fix paths for cross-platforms
    #
    # @param path   the path to fix
    # @return the fixed path
    @classmethod
    def fix_path(cls, path):
        path = os.path.expanduser(path)
        if svc.gbl.os_name == 'win':
            # assumes there is only one ":" and it is for a drive letter
            m = re.search(r'(.*)(.):(.*)', path)
            if m:
                drive = m.group(2).lower()
                path = f'{m.group(1)}/{drive}/{m.group(3)}'
        path = path.replace('\\', '/')
        path = path.replace('//', '/')
        return path

    # --------------------
    ## get the homebrew link libraries root directory
    #
    # @return the homebrew link lib root dir
    @classmethod
    def homebrew_link_dirs(cls):
        return []  # TODO del; '/opt/homebrew/lib'

    # --------------------
    ## get the homebrew includes root directory
    #
    # @return the homebrew includes dir
    @classmethod
    def homebrew_inc_dirs(cls):
        return [
            # TODO del;
            # '/opt/homebrew/opt/llvm/include',
            # '/opt/homebrew/opt/gcc/include/c++/15',
            # '/opt/homebrew/Cellar/gcc/15.1.0',
            # '/opt/homebrew/include',
        ]

    # --------------------
    ## get C++ compiler per OS
    # @return C++ compiler
    @classmethod
    def cpp_compiler(cls):
        if svc.gbl.os_name == 'macos':
            comp = '/opt/homebrew/Cellar/gcc/15.1.0/bin/g++-15'
        else:
            comp = 'g++'
        return comp

    # --------------------
    ## get C compiler per OS
    # @return C compiler
    @classmethod
    def c_compiler(cls):
        if svc.gbl.os_name == 'macos':
            comp = '/opt/homebrew/Cellar/gcc/15.1.0/bin/gcc-15'
        else:
            comp = 'gcc'
        return comp

    # --------------------
    ## get the root of the arduino cores, tools, etc.
    #
    # @return the root arduino directory
    @classmethod
    def arduino_root_dir(cls):
        if svc.gbl.os_name == 'macos':
            path = os.path.expanduser('~/Library/Arduino15')
        elif svc.gbl.os_name == 'win':
            path = os.path.expanduser('~/AppData/Local/Arduino15')
        else:
            path = '/usr/share/arduino'
        # do not fix_path()
        return path

    # --------------------
    ## get the library directory for arduino core source files
    #
    # @param arduino_root_dir   the root of the arduino directory system
    # @return the root libdir
    @classmethod
    def arduino_core_libdir(cls, arduino_root_dir):
        if svc.gbl.os_name == 'macos':
            path = f'{arduino_root_dir}/packages/arduino/hardware/avr/1.8.6/cores/arduino'
        elif svc.gbl.os_name == 'win':
            path = f'{arduino_root_dir}/packages/arduino/hardware/avr/1.8.6/cores/arduino'
        else:
            path = f'{arduino_root_dir}/hardware/arduino/avr/cores/arduino'
        # do not fix_path()
        return path

    # --------------------
    ## get the list of included directories for arduino core
    #
    # @param arduino_root_dir   the root of the arduino directory system
    # @return the list of include directories
    @classmethod
    def arduino_core_includes(cls, arduino_root_dir):
        if svc.gbl.os_name == 'macos':
            incs = [
                f'{arduino_root_dir}/packages/arduino/hardware/avr/1.8.6/cores/arduino',
                f'{arduino_root_dir}/packages/arduino/hardware/avr/1.8.6/variants/standard',
            ]
        elif svc.gbl.os_name == 'win':
            incs = [
                f'{arduino_root_dir}/packages/arduino/hardware/avr/1.8.6/cores/arduino',
                f'{arduino_root_dir}/packages/arduino/hardware/avr/1.8.6/variants/standard',
            ]
        else:
            incs = [
                f'{arduino_root_dir}/hardware/arduino/avr/cores/arduino',
                f'{arduino_root_dir}/hardware/arduino/avr/variants/standard',
            ]
        # do not fix_path()
        return incs

    # --------------------
    ## get the directory for avrdude.conf
    #
    # @param arduino_root_dir   the root of the arduino directory system
    # @return the avrdude directory
    @classmethod
    def avrdude_dir(cls, arduino_root_dir):
        if svc.gbl.os_name == 'macos':
            path = '/opt/homebrew/etc'
        elif svc.gbl.os_name == 'win':
            path = f'{arduino_root_dir}/packages/arduino/tools/avrdude/6.3.0-arduino17/etc'
            # TODO /c/Users/micro/AppData/Local/Arduino15/packages/arduino/tools/avrdude/6.3.0-arduino17/etc
        else:
            path = f'{arduino_root_dir}/hardware/tools'
        # do not fix_path()
        return path

    # --------------------
    ## return default gtest include directories
    #
    # @return the list of include directories
    @classmethod
    def gtest_includes(cls):
        incs = []
        if svc.gbl.os_name == 'win':
            incs.append('c:/msys64/mingw64/include')
        elif svc.gbl.os_name == 'macos':
            incs.append('/opt/homebrew/Cellar/googletest/1.17.0/include')

        # do not fix_path()
        return incs

    # --------------------
    ## return default gtest link directories
    #
    # @return the list of link directories
    @classmethod
    def gtest_link_dirs(cls):
        dirs = []
        if svc.gbl.os_name == 'win':
            dirs.append('c:/msys64/mingw64/lib')
        elif svc.gbl.os_name == 'macos':
            dirs.append('/opt/homebrew/Cellar/googletest/1.17.0/lib')
        # do not fix_path()
        return dirs

    # --------------------
    ## return default ruby include directories
    #
    # @return the list of include directories
    @classmethod
    def ruby_includes(cls):
        # TODO del; check if these work in msys2
        # incs = []
        # if svc.gbl.os_name == 'win':
        #     incs.append('C:/Ruby33-x64/include/ruby-3.3.0')
        #     incs.append('C:/Ruby33-x64/include/ruby-3.3.0/x64-mingw-ucrt')

        incs = [
            cls._get_ruby_inc1(),
            cls._get_ruby_inc2()
        ]
        # uncomment to debug
        # print(f'@@@DBG inc1:{incs[0]}')
        # print(f'@@@DBG inc2:{incs[1]}')

        # do not fix_path()
        return incs

    # --------------------
    ## get include directory1 using ruby utility
    #
    # @return path
    @classmethod
    def _get_ruby_inc1(cls):
        cmd = ['ruby', '-rrbconfig', '-e', 'puts RbConfig::CONFIG["rubyhdrdir"]']
        result = subprocess.run(cmd, capture_output=True, text=True, check=True, shell=False)
        lines = result.stdout.strip().splitlines()
        if lines:
            inc1 = lines[0].strip()
        else:
            if svc.gbl.os_name == 'macos':
                inc1 = '/opt/homebrew/Cellar/ruby/3.3.5/include/ruby-3.3.0'
            elif svc.gbl.os_name == 'ubuntu':
                inc1 = '/usr/include/ruby-3.2.0'
            else:
                inc1 = 'unknown_inc1'
        return inc1

    # --------------------
    ## get include directory2 using ruby utility for ubuntu
    #
    # @return path
    @classmethod
    def _get_ruby_inc2(cls):
        cmd = ['ruby', '-rrbconfig', '-e', 'puts RbConfig::CONFIG["rubyarchhdrdir"]']
        result = subprocess.run(cmd, capture_output=True, text=True, check=True, shell=False)
        lines = result.stdout.strip().splitlines()
        if lines:
            inc2 = lines[0].strip()
        else:
            if svc.gbl.os_name == 'macos':
                inc2 = '/opt/homebrew/Cellar/ruby/3.3.5/include/ruby-3.3.0'
            elif svc.gbl.os_name == 'ubuntu':
                inc2 = '/usr/include/x86_64-linux-gnu/ruby-3.2.0'
            else:
                inc2 = 'unknown_inc2'
        return inc2

    # --------------------
    ## return default ruby link directories
    #
    # @return the list of link directories
    @classmethod
    def ruby_link_dirs(cls):
        dirs = [cls._get_ruby_lib()]

        # uncomment to debug
        # print(f'@@@DBG link:{dirs[0]}')

        # do not fix_path()
        return dirs

    # --------------------
    ## get link library using ruby utility
    #
    # @return link library
    @classmethod
    def _get_ruby_lib(cls):
        # TODO check if this works in msys2
        cmd = ['ruby', '-rrbconfig', '-e', 'puts RbConfig::CONFIG["libdir"]']
        result = subprocess.run(cmd, capture_output=True, text=True, check=True, shell=False)
        lines = result.stdout.strip().splitlines()
        if lines:
            lib = lines[0].strip()
        elif svc.gbl.os_name == 'macos':
            # best guess
            lib = '/opt/homebrew/Cellar/ruby/3.3.5/lib'
        elif svc.gbl.os_name == 'win':
            lib = 'C:/Ruby33-x64/lib'
        else:  # ubuntu
            lib = '/usr/lib'

        # uncomment to debug
        # print(f'@@@DBG lib:{lib}')
        return lib

    # --------------------
    ## return default link libraries needed for ruby
    #
    # @return the list of link libraries
    @classmethod
    def ruby_link_libs(cls):
        libs = []
        if svc.gbl.os_name == 'win':
            libs.append('x64-ucrt-ruby330.dll')
        # do not fix_path()
        return libs

    # --------------------
    ## return default python include directories
    #
    # @return the list of include directories
    @classmethod
    def python_includes(cls):
        # python3 -c "import sysconfig; print(sysconfig.get_path('include'))"
        import sysconfig
        incs = [sysconfig.get_path('include')]
        # do not fix_path()
        return incs

    # --------------------
    ## return default link libraries needed for ruby
    #
    # @return the list of link libraries
    @classmethod
    def python_link_libs(cls):
        import sysconfig
        libs = [sysconfig.get_config_var('LIBDIR')]

        # TODO del;
        # libs = []
        # if svc.gbl.os_name == 'macos':
        #     libs.append(
        #         '/opt/homebrew/Cellar/python@3.10/3.10.15/Frameworks/Python.framework/Versions/3.10/lib/python3.10/config-3.10-darwin')
        # do not fix_path()
        return libs



# --------------------
## holds singletons and global values
class svc:  # pylint: disable=invalid-name
    ## holds reference to logger
    log = None

    ## holds global information
    gbl = None

    ## holds OS Abstraction Layer
    osal = None

    # --------------------
    ## abort the current script.
    # Note: do not use logging here since it may fail to write correctly
    #
    # @param msg  (optional) message to display
    # @return does not return
    @classmethod
    def abort(cls, msg='abort occurred, exiting'):
        import sys
        print('')
        print(f'ABRT {msg}')
        sys.stdout.flush()
        sys.exit(1)



from .svc import svc
from .target_c_cpp_lib_base import TargetCCppLibBase


# --------------------
## generate a C Library target, static (default) or shared
class TargetCLib(TargetCCppLibBase):
    # --------------------
    ## create a C++ library instance
    #
    # @param targets      current list of targets
    # @param target_name  name of new target to add
    @classmethod
    def create(cls, targets, target_name):
        impl = TargetCLib(target_name)
        targets.append(impl)
        return impl

    # --------------------
    ## constructor
    #
    # @param target_name  the name of this target
    def __init__(self, target_name):
        super().__init__(target_name)

        ## the C lib target type
        self._target_type = 'c-lib'
        ## list of object files
        self._objs = ''
        ## compiler to use
        self._cxx = svc.osal.c_compiler()
        ## list of compile options
        self.add_compile_options('-std=c17')  # pylint: disable=E1101

        ## list of build directories
        self._build_dirs = {}

        ## library type: static, shared
        self._lib_type = 'static'

    # --------------------
    ## return target type
    #
    # @return cpp target
    @property
    def target_type(self):
        return self._target_type

    # --------------------
    ## return compiler to use
    # @return compiler to use
    @property
    def compiler(self):
        return self._cxx

    # --------------------
    ## set compiler to use
    # @param val  the compiler setting
    # @return None
    @compiler.setter
    def compiler(self, val):
        self._cxx = val
        # not needed, but adding anyway
        self._compile_opts_param.remove('-std=c++20')  # pylint: disable=E1101



import json
import os

from .svc import svc


# --------------------
## Shared info for Arduino Core targets and arduino targets
class ArduinoShared:
    ## board id of the arduino; set by user
    boardid = None
    ## avrdude port to use
    avrdude_port = None

    # these are from /usr/share/arduino/hardware/arduino/avr/boards.txt
    ## the CPU id
    f_cpu = None
    ## the MCU id
    mcu = None
    ## the avrdude to use
    avrdude = None
    ## the avrdude baud rate to use for this board
    avrdude_baudrate = None
    ## the avrdude protocol to use for this board
    avrdude_protocol = None
    # these are derived from the above
    ## common compile options
    common_opts = None
    ## C++ compile options
    cpp_opts = None
    ## C compile options
    cc_opts = None

    ## debug compile options
    debug_compile_opts = None

    ## arduino directory that holds common files and arduino library
    arduino_dir = None
    ## arduino AVR C++ compiler
    cpp = None
    ## arduino AVR C compiler
    cc = None
    ## arduino AVR linker
    ar = None
    ## arduino copies an object file
    obj_copy = None
    ## path to avrdude
    avrdude_dir = None

    # core related
    ## name of core target
    core_tgt = None
    ## path to core lib
    corelib = None
    ## name of the core lib
    corelib_name = None
    ## the path to the core lib
    coredir = None
    ## additional include directories for the arduino library
    core_includes = []

    ## holds library root directory
    library_root = ''

    # --------------------
    ## constructor
    def __init__(self):
        ## debug compile options
        self.debug_compile_opts = ['-g', '-Os', '-Wall']

        ## arduino directory that holds common files and arduino library
        self.arduino_dir = svc.osal.arduino_root_dir()
        ## arduino AVR C++ compiler
        self.cpp = 'avr-g++'
        ## arduino AVR C compiler
        self.cc = 'avr-gcc'
        ## arduino AVR linker
        self.ar = 'avr-ar'
        ## arduino copies an object file
        self.obj_copy = 'avr-objcopy'
        ## path to avrdude.conf
        self.avrdude_dir = svc.osal.avrdude_dir(self.arduino_dir)

        ## additional include directories for the arduino library
        self.core_includes = svc.osal.arduino_core_includes(self.arduino_dir)

    # --------------------
    ## set location of the root directory
    #
    # @param rootdir  the path to the root dir
    # @return None
    def set_library_root(self, rootdir):
        ## see ArduinoShared.Library_root
        self.library_root = rootdir

    # --------------------
    ## set the avrdude port
    #
    # @param val   the port to use
    # @return None
    def set_avrdude_port(self, val):
        self.avrdude_port = val
        self._check_port_exists(val)

    # --------------------
    ## print the list of known Arduino boards
    #
    # @return None
    def print_board_list(self):
        boards = self._get_board_json()
        svc.log.line('Available boards:')
        for name, info in boards.items():
            svc.log.line(f'   {name: <20}: {info["fullname"]}')

    # --------------------
    ## check common arduino values
    #
    # @return None
    def check(self):
        errs = 0
        errs = self._check_arg(errs, 'boardid')
        errs = self._check_arg(errs, 'avrdude_port')
        errs = self._check_arg(errs, 'f_cpu')
        errs = self._check_arg(errs, 'mcu')
        errs = self._check_arg(errs, 'avrdude')
        errs = self._check_arg(errs, 'avrdude_baudrate')
        errs = self._check_arg(errs, 'avrdude_protocol')
        errs = self._check_arg(errs, 'common_opts')
        errs = self._check_arg(errs, 'cpp_opts')
        errs = self._check_arg(errs, 'cc_opts')

        errs = self._check_arg(errs, 'core_tgt')
        errs = self._check_arg(errs, 'coredir')
        errs = self._check_arg(errs, 'corelib')
        errs = self._check_arg(errs, 'corelib_name')

        if errs > 0:
            svc.abort('arduino: resolve errors')

    # --------------------
    ## check argument for arduino board
    #
    # @param errs   current number of errs
    # @param arg    the argument to check
    # @return the updated errs value
    def _check_arg(self, errs, arg):
        selfarg = getattr(self, arg, None)
        if selfarg is None:
            errs += 1
            svc.log.err(f'arduino: {arg} is not set')
        return errs

    # --------------------
    ## set board id for this shared component
    #
    # @param boardid the board id
    # @return None
    def set_boardid(self, boardid):
        info = self._get_board_info(boardid)
        self.boardid = boardid

        # set values based on board id
        self.f_cpu = info['build.f_cpu']
        self.mcu = info['build.mcu']
        self.avrdude = info['upload.tool']
        self.avrdude_baudrate = info['upload.speed']
        self.avrdude_protocol = info['upload.protocol']

        self.common_opts = f'-MMD -c -ffunction-sections -fdata-sections ' \
                           f'-mmcu={self.mcu} -DF_CPU={self.f_cpu}L ' \
                           '-DUSB_VID=null -DUSB_PID=null -DARDUINO=106'
        self.cpp_opts = f'{self.common_opts} -fno-exceptions -fno-threadsafe-statics -std=c++11'
        self.cc_opts = self.common_opts

    # --------------------
    ## get board info based on the boardid
    #
    # @param boardid  the board id to search for
    # @return info found
    def _get_board_info(self, boardid):
        boards = self._get_board_json()
        if boardid not in boards:
            svc.abort(f'arduino: invalid boardid: {boardid}')

        info = boards[boardid]
        return info

    # --------------------
    ## load board info from the boards.json file
    #
    # @return the board info
    def _get_board_json(self):
        path = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'boards.json')
        with open(path, 'r', encoding='utf-8') as fp:
            boards = json.load(fp)
            return boards

    # --------------------
    ## check if a serial port exists
    #
    # @param port   the port to check
    # @return None
    def _check_port_exists(self, port):
        try:
            import serial
        except ModuleNotFoundError:
            svc.log.warn(f'{"arduino_shared": <15}: module pyserial is not installed')
            return

        import serial.tools.list_ports
        ports = serial.tools.list_ports.comports()
        foundit = False
        for found_port, desc, hwid in sorted(ports):
            if port == found_port:
                foundit = True
                svc.log.line(f'{"arduino_shared": <15}: found port {port}: {desc} [{hwid}]')
                break
        if not foundit:
            svc.log.warn(f'{"arduino_shared": <15}: port {port} not found')



from dataclasses import dataclass

from .svc import svc


# --------------------
## info for the OpenGL package
class PackageOpengl:
    # --------------------
    ## return info about the pkgname
    #
    # @param pkgname  the name of the package; default: opengl
    # @return ref to OpenGl info
    def find(self, pkgname):
        svc.log.line(f'finding package: {pkgname}')

        @dataclass
        class OpenGl:
            include_dir = '/usr/include'
            link_libs = ['glut', 'GLU', 'GL']

        return OpenGl



import os

from .svc import svc
from .target_base import TargetBase


# --------------------
## generate a gtest target
class TargetGtest(TargetBase):
    # --------------------
    ## create a gtest target instance
    #
    # @param targets      current list of targets
    # @param target_name  name of new target to add
    @classmethod
    def create(cls, targets, target_name):
        impl = TargetGtest(target_name)
        targets.append(impl)
        return impl

    # --------------------
    ## constructor
    #
    # @param target_name  the name of this target
    def __init__(self, target_name):
        super().__init__(target_name)

        ## list of object files
        self._objs = ''
        ## compiler to use
        self._cxx = 'g++'

        ## list of compile options
        self.add_compile_options([  # pylint: disable=E1101
            '-g', '-fdiagnostics-color=always', '-fprofile-arcs',
            '-ftest-coverage', '-DGTEST_HAS_PTHREAD=1', '-std=gnu++20',
            '-D_UCRT', '-D_GNU_SOURCE',
        ])

        ## list of build directories
        self._build_dirs = {}

        ## list of include directories
        self.add_include_directories(svc.osal.gtest_includes())  # pylint: disable=E1101

        ## list of link libraries to add
        self.add_link_libraries(['gtest', 'pthread'])  # pylint: disable=E1101
        if svc.gbl.os_name != 'macos':
            self.add_link_libraries('gcov')  # pylint: disable=E1101

        ## list of link directories to add
        self.add_link_directories(svc.osal.gtest_link_dirs())  # pylint: disable=E1101

        if svc.gbl.os_name == 'macos':
            self.add_link_options('--coverage')

        ## list of coverage directories
        self._cov_dirs = []

    # --------------------
    ## return target type
    #
    # @return gtest target
    @property
    def target_type(self):
        return 'gtest'

    # --------------------
    ## add coverage directories to list to cover
    #
    # @param cov_list   list of directories to add
    # @return None
    def add_coverage(self, cov_list):
        if isinstance(cov_list, list):
            pass
        elif isinstance(cov_list, str):
            # convert to a list
            cov_list = [cov_list]
        else:
            svc.abort('add_coverage(): accepts only str or list of str')

        for cov_dir in cov_list:
            if not isinstance(cov_dir, str):
                svc.abort(f'add_coverage(): accepts only str or list of str, {cov_dir} is {type(cov_dir)}')

            # user can add an empty entry
            if cov_dir == '':
                continue

            self._cov_dirs.append(cov_dir)

    # --------------------
    ## check target for any issues
    #
    # @return None
    def check(self):
        svc.log.highlight(f'{self.target}: check target...')
        self._common_check()

        for covdir in self._cov_dirs:
            if not os.path.isdir(covdir):
                svc.log.warn(f'{self.target}: coverage directory {covdir} not found')

    # --------------------
    ## gen gtest target
    #
    # @return None
    def gen_target(self):
        svc.log.highlight(f'{self.target}: gen target, type:{self.target_type}')

        self._gen_args()
        self._gen_init()
        self._gen_app()
        self._gen_link()
        self._gen_coverage()
        self._gen_run()

    # --------------------
    ## create output directory
    #
    # @return None
    def _gen_args(self):
        # create output build directory
        self._build_dirs[svc.gbl.build_dir] = 1

        for file in self.sources:  # pylint: disable=E1101
            _, _, dst_dir = self._get_obj_path(file)
            self._build_dirs[dst_dir] = 1

        self._writeln('')

    # --------------------
    ## gen initial content for gtest target
    #
    # @return None
    def _gen_init(self):
        rule = f'{self.target}-init'
        self.add_rule(rule)

        self._gen_rule(rule, '', f'{self.target}: initialize for {svc.gbl.build_dir} build')
        for blddir in self._build_dirs:
            self._writeln(f'\t@mkdir -p {svc.osal.fix_path(blddir)}')
        self._writeln('')

    # --------------------
    ## gen coverage to clean patterns list
    #
    # @param pattern  directory or pattern to add to list to clean
    # @return None
    def _add_clean_cov(self, pattern):
        if pattern not in self._clean_cov:
            self._clean_cov[pattern] = 1

    # --------------------
    ## gen app build target
    #
    # @return None
    def _gen_app(self):
        rule = f'{self.target}-build'
        self.add_rule(rule)

        build_deps = ''
        for file in self.sources:  # pylint: disable=E1101
            obj, mmd_inc, dst_dir = self._get_obj_path(file)

            # gen clean paths
            clean_path = svc.osal.fix_path(dst_dir.replace(f'{svc.gbl.build_dir}/', ''))
            self.add_clean(f'{clean_path}/*.o')
            self.add_clean(f'{clean_path}/*.d')
            # coverage related cleans
            self.add_clean(f'{clean_path}/*.gcno')
            self._add_clean_cov(f'{clean_path}/*.gcda')

            self._writeln(f'-include {svc.osal.fix_path(mmd_inc)}')
            self._writeln(f'{svc.osal.fix_path(obj)}: {svc.osal.fix_path(file)}')
            self._writeln(f'\t{self._cxx} -MMD -c {self._inc_dirs} {self._compile_opts} '
                          f'{svc.osal.fix_path(file)} -o {svc.osal.fix_path(obj)}')
            self._objs += f'{svc.osal.fix_path(obj)} '
            build_deps += f'{svc.osal.fix_path(file)} '
        self._writeln('')

        self._gen_rule(rule, self._objs, f'{self.target}: build source files')
        self._writeln('')

    # --------------------
    ## gen link target
    #
    # @return None
    def _gen_link(self):
        rule = f'{self.target}-link'
        self.add_rule(rule)

        exe = f'{svc.gbl.build_dir}/{self.target}'
        self._writeln(f'{exe}: {self._objs}')
        self._writeln(f'\t{self._cxx} {self._objs} {self._link_opts} {self._link_paths} {self._libs} -o {exe}')
        self._writeln('')
        ## see baseclass for definition of self.target
        self.add_clean(self.target)

        self._gen_rule(rule, f'{exe} {self.target}-build', f'{self.target}: link')
        self._writeln('')

    # --------------------
    ## gen coverage target
    #
    # @return None
    def _gen_coverage(self):
        rule = f'{self.target}-cov'
        # don't add to rules

        report_page = f'{svc.gbl.build_dir}/{self.target}.html'

        cmd = ('gcovr --html-details '  # show individual source files
               # '--no-color '  # no color for text output; not used in Ubuntu 24.04 version
               '--sort=uncovered-percent '  # sort source files based on percentage uncovered lines
               '--print-summary '  # print summary to stdout
               f'-o {report_page} ')  # location of report main page

        if not self._cov_dirs:
            svc.log.warn('gen_coverage: cov_dirs is empty, use add_coverage()')

        for cov_dir in self._cov_dirs:
            cmd += f'--filter {cov_dir} '

        self.add_clean(f'{self.target}.html')
        self.add_clean(f'{self.target}.css')
        self.add_clean(f'{self.target}.**.html')

        self._gen_rule(rule, '', f'{self.target}: show coverage')
        self._writeln(f'\t{cmd}')
        self._writeln(f'\t@echo "see {report_page}" for HTML report')
        self._writeln('')

    # --------------------
    ## gen target to run gtest target
    #
    # @return None
    def _gen_run(self):
        rule = f'{self.target}-run'
        # don't add rule

        self._gen_empty_rule()
        exe = f'{svc.gbl.build_dir}/{self.target}'
        self._gen_rule(rule, f'{self.target}-link', f'{self.target}: run executable')
        self._writeln(f'\t{exe} $(filter-out $@,$(MAKECMDGOALS))')
        self._writeln('')



from .svc import svc
from .target_base import TargetBase


# --------------------
## generate a C++ Library target, static (default) or shared
# see https://renenyffenegger.ch/notes/development/languages/C-C-plus-plus/GCC/create-libraries/index
class TargetCCppLibBase(TargetBase):
    # create() is c/c++ specific

    # --------------------
    ## constructor
    #
    # @param target_name  the name of this target
    def __init__(self, target_name):
        super().__init__(target_name)

        ## the C/C++ lib target type
        self._target_type = 'unset'
        ## list of object files
        self._objs = ''
        ## compiler to use
        self._cxx = 'unset'
        ## list of compile options
        self.add_compile_options('-D_GNU_SOURCE')  # pylint: disable=E1101

        ## list of build directories
        self._build_dirs = {}

        ## library type: static, shared
        self._lib_type = 'static'

    # target_type() is c/c++ specific
    # compiler() getter/setter are c/c++ specific

    # --------------------
    ## set the library type to generate
    #
    # @param val  the type to set: static or shared
    def set_type(self, val):
        valid_types = ['static', 'shared']
        if val not in valid_types:
            svc.abort(f'invalid library type, expected {valid_types}, actual: {val}')
        self._lib_type = val

    # --------------------
    ## check target for any issues
    #
    # @return None
    def check(self):
        svc.log.highlight(f'{self.target}: check target...')
        self._common_check()

    # --------------------
    ## gen C++ library target
    #
    # @return None
    def gen_target(self):
        svc.log.highlight(f'{self.target}: gen target, type:{self._target_type}')

        self._gen_args()
        self._gen_init()
        self._gen_lib()
        if self._lib_type == 'shared':
            self._gen_shared_library()
        else:
            self._gen_static_library()

    # --------------------
    ## create output directory
    #
    # @return None
    def _gen_args(self):
        # create output build directory
        self._build_dirs[svc.gbl.build_dir] = 1

        for file in self.sources:  # pylint: disable=E1101
            _, _, dst_dir = self._get_obj_path(file)
            self._build_dirs[dst_dir] = 1

        self._writeln('')

    # --------------------
    ## gen initial content for C++ library
    #
    # @return None
    def _gen_init(self):
        rule = f'{self.target}-init'
        self.add_rule(rule)

        self._gen_rule(rule, '', f'{self.target}: initialize for {svc.gbl.build_dir} build')
        for bld_dir in self._build_dirs:
            self._writeln(f'\t@mkdir -p {svc.osal.fix_path(bld_dir)}')
        self._writeln('')

    # --------------------
    ## gen lib build target
    #
    # @return None
    def _gen_lib(self):
        rule = f'{self.target}-build'
        self.add_rule(rule)

        build_deps = ''
        for file in self.sources:  # pylint: disable=E1101
            obj, mmd_inc, dst_dir = self._get_obj_path(file)

            # gen clean paths
            clean_path = dst_dir.replace(f'{svc.gbl.build_dir}/', '')
            self.add_clean(f'{svc.osal.fix_path(clean_path)}/*.o')
            self.add_clean(f'{svc.osal.fix_path(clean_path)}/*.d')

            self._writeln(f'-include {svc.osal.fix_path(mmd_inc)}')
            self._writeln(f'{svc.osal.fix_path(obj)}: {svc.osal.fix_path(file)}')
            fpic = ''
            if self._lib_type == 'shared':
                fpic = '-fPIC '
            self._writeln(f'\t{self._cxx} -MMD {fpic} -c {self._inc_dirs} {self._compile_opts} '
                          f'{svc.osal.fix_path(file)} -o {svc.osal.fix_path(obj)}')
            self._objs += f'{svc.osal.fix_path(obj)} '
            build_deps += f'{svc.osal.fix_path(file)} '

        self._writeln('')

        self._gen_rule(rule, self._objs, f'{self.target}: build source files')
        self._writeln('')

    # --------------------
    ## gen shared library
    #
    # @return None
    def _gen_shared_library(self):
        rule = f'{self.target}-shared'
        self.add_rule(rule)

        if svc.gbl.os_name == 'win':
            extension = 'dll'
        else:
            extension = 'so'
        lib_name = f'lib{self.target}.{extension}'
        lib = f'{svc.gbl.build_dir}/{lib_name}'
        self._writeln(f'{svc.osal.fix_path(lib)}: {self._objs}')
        self._writeln(
            f'\t{self._cxx} -MMD --shared -fPIC {self._objs} {self._link_opts} {self._link_paths} {self._libs} '
            f'-o {svc.osal.fix_path(lib)}')
        self._writeln('')
        ## see baseclass for definition of self.target
        self.add_clean(lib_name)

        self._gen_rule(rule, lib, f'{self.target}: link')
        self._writeln('')

    # --------------------
    ## gen static library
    #
    # @return None
    def _gen_static_library(self):
        rule = f'{self.target}-shared'
        self.add_rule(rule)

        lib_name = f'lib{self.target}.a'
        lib = f'{svc.gbl.build_dir}/{lib_name}'
        self._writeln(f'{svc.osal.fix_path(lib)}: {self._objs}')
        self._writeln(
            f'\tar rcs {svc.osal.fix_path(lib)} {self._objs} {self._link_opts} {self._link_paths} {self._libs}')
        self._writeln('')
        ## see baseclass for definition of self.target
        self.add_clean(lib_name)

        self._gen_rule(rule, lib, f'{self.target}: link')
        self._writeln('')



from .svc import svc
from .target_base_min import TargetBaseMin


# --------------------
## generate a target with given commands, help etc.
class TargetManual(TargetBaseMin):
    # class TargetManual(TargetBase):
    # --------------------
    ## create a manual target instance
    #
    # @param targets      current list of targets
    # @param target_name  name of new target to add
    @classmethod
    def create(cls, targets, target_name):
        impl = TargetManual(target_name)
        targets.append(impl)
        return impl

    # --------------------
    ## constructor
    #
    # @param target_name  the name of this target
    def __init__(self, target_name):
        super().__init__(target_name)

        ## commands to invoke
        self._cmds = []

        ## dependencies for this target
        self._deps = []

        ## help line to display
        self._help_line = 'run commands'

    # --------------------
    ## return target type
    #
    # @return manual target
    @property
    def target_type(self):
        return 'manual'

    # --------------------
    ## add a command to this target
    #
    # @param line   the command line to add
    # @return None
    def add_command(self, line):
        self._cmds.append(line)

    # --------------------
    ## add a dependency to this target
    #
    # @param dependency   name of another rule to invoke before this target
    # @return None
    def add_dependency(self, dependency):
        self._deps.append(dependency)

    # --------------------
    ## set the description displayed in the help
    #
    # @param line   the help to use
    # @return None
    def set_help(self, line):
        self._help_line = line

    # --------------------
    ## check target for any issues
    #
    # @return None
    def check(self):
        # nothing to check
        pass

    # --------------------
    ## gen Manual target
    #
    # @return None
    def gen_target(self):
        svc.log.highlight(f'{self.target}: gen target, type:{self.target_type}')

        self._gen_run()

    # --------------------
    ## gen target to run the manual command
    #
    # @return None
    def _gen_run(self):
        deps = ''
        if len(self._deps) > 0:
            deps = ' '.join(self._deps)

        self._gen_rule(self.target, deps, f'{self._help_line}')
        for cmd in self._cmds:
            self._writeln(f'\t{cmd}')
        self._writeln('')



import os

from .svc import svc


# --------------------
## handle CPIP packages
class PackageCpip:
    # --------------------
    ## find the CPIP package name and return info for it
    #
    # @param pkgname  the CPIP package to search for
    # @return the package info
    def find(self, pkgname):
        pkgname = pkgname.replace('cpip.', '')
        svc.log.line(f'finding cpip package: {pkgname}')

        if not os.path.isdir(os.path.join('tools', 'xplat_utils')):
            svc.abort('xplat_utils is not installed, aborting')

        from tools.xplat_utils import main
        pkginfo = main.svc.utils.cpip_get(pkgname)
        if pkginfo is None:
            svc.abort(f'could not find info for cpip package: {pkgname}')

        return pkginfo



import os

from .svc import svc


# --------------------
## List oriented parameter
class ListParam:
    # --------------------
    ## constructor
    #
    # @param tag         the function name tag
    # @param parm_type   list of strings or paths
    # @param parent      the parent class
    # @param filter_fn   optional filter function callback; used to filter out various items; otherwise None
    # @param accum_fn    optional accumulator function callback; gathers the list of items into a parameter string
    def __init__(self, tag, parm_type, parent, filter_fn, accum_fn):
        ## holds the name of parameter; to generate function names
        self._tag = tag
        ## parameter type: either 'string' or 'path'
        self._parm_type = parm_type
        ## reference to the parent class
        self._parent = parent
        ## holds the parent class's name
        self._parent_name = self._parent.__class__.__name__
        ## filter callback function
        self._filter_fn = filter_fn
        ## accumulator callback function
        self._accum_fn = accum_fn

        if self._parm_type not in ['string', 'path']:
            svc.abort(f'ListParm {tag}: unknown type "{parm_type}"')

        ## holds the current list of parameter values
        self._values = []

        # define function (property) to return current values
        ## see property values; note will overwrite any existing function
        setattr(self._parent, self._tag, self.values)

        # define function to add values; note will overwrite any existing function
        fn = f'add_{self._tag}'
        ## see function update()
        setattr(self._parent, fn, self.update)

        # define function to remove a single value; note will overwrite any existing function
        fn = f'remove_{self._tag}'
        ## see function update()
        setattr(self._parent, fn, self.remove)

    # --------------------
    ## returns list of current values
    # @return list of current values
    @property
    def values(self):
        return self._values

    # --------------------
    ## remove the given value, if present in the list of values
    #
    # @param vals  the values to remove
    # @return None
    def remove(self, vals):
        if isinstance(vals, list):
            pass
        elif isinstance(vals, str):
            # convert to a list
            vals = [vals]
        else:
            svc.abort(f'{self._tag}: can only remove strings: {vals} is {type(vals)}')

        for val in vals:
            if not isinstance(val, str):
                svc.abort(f'{self._tag}: accepts only str or list of str, {val} is {type(val)}')

            if val in self._values:
                self._values.remove(val)
                if self._accum_fn:
                    self._accum_fn()

    # --------------------
    ## add additional values to the current list.
    #   * skips if the value is already present
    #   * skips if the value is empty "" or None
    #   * aborts if any of the values passed in is not a string
    #
    # @param vals  the values to add; accepts a string or a list of strings
    # @return None
    def update(self, vals):
        if isinstance(vals, list):
            pass
        elif isinstance(vals, str):
            # convert to a list
            vals = [vals]
        else:
            svc.abort(f'{self._tag}: can only add strings: {vals} is {type(vals)}')

        for val in vals:
            if not isinstance(val, str):
                svc.abort(f'{self._tag}: accepts only str or list of str, {val} is {type(val)}')

            # user can add an empty entry
            if not val:
                continue

            # if value matches filter, then don't add
            if self._filter_fn and self._filter_fn(val):
                continue

            if self._parm_type == 'path':
                updated_val = os.path.expanduser(val)
            else:  # parm_type == 'string'
                updated_val = val
            updated_val = os.path.normpath(updated_val)

            if updated_val in self._values:
                svc.log.line(f'{self._tag}: {val} is already added: {self._values}')
                continue

            self._values.append(updated_val)

        if self._accum_fn:
            self._accum_fn()



from .svc import svc
from .target_base import TargetBase


# --------------------
## target for an Arduino app
class TargetArduino(TargetBase):
    # --------------------
    ## create an Arduino target
    #
    # @param targets      current list of targets
    # @param target_name  name of the new target
    # @param shared       info for shared core
    @classmethod
    def create(cls, targets, target_name, shared=None):
        impl = TargetArduino(target_name, shared)
        targets.append(impl)
        return impl

    # --------------------
    ## constructor
    #
    # @param target_name  name of the new target
    # @param shared       info for shared core
    def __init__(self, target_name, shared):
        super().__init__(target_name)

        if shared is None:
            svc.abort('Arduino Targets must have a shared Arduino Core component passed in')

        ## shared Arduino Core info
        self._shared = shared

        ## compile options
        self.add_compile_options(self._shared.debug_compile_opts)  # pylint: disable=E1101

        ## build directories
        self._build_dirs = {}

        # these need to be set below
        ## target build directory
        self.tgt_build_dir = None
        ## ELF file generated for this app
        self._elffile = None
        ## EEP file generated for this app
        self._eepfile = None
        ## HEX file generated for this app
        self._hexfile = None
        ## list of target dependencies for this app
        self._app_deps = None

    # --------------------
    ## type of this target
    # @return arduino target
    @property
    def target_type(self):
        return 'arduino'

    # --------------------
    ## update include directories params string
    #
    # @return None
    def _update_inc_dirs(self):
        super()._update_inc_dirs()
        for incdir in self._shared.core_includes:
            self._inc_dirs += f' "-I{svc.osal.fix_path(incdir)}" '

    # --------------------
    ## add an arduino library to list of libraries.
    # these libraries are recognized:
    #  * Servo
    #
    # @param lib  the library to add
    # @return None
    def add_library(self, lib):
        if lib == 'Servo':
            svc.log.line(f'adding arduino library: {lib}')
            self.add_sources(f'{self._shared.library_root}/Servo/src/avr/Servo.cpp')  # pylint: disable=E1101
            self.add_include_directories(f'{self._shared.library_root}/Servo/src')  # pylint: disable=E1101
            self.add_compile_options('-DARDUINO_ARCH_AVR')  # pylint: disable=E1101
        else:
            svc.abort(f'{self.target} target_arduino: unknown arduino library: {lib}')

    # --------------------
    ## check various aspects of parameters
    #
    # @return None
    def check(self):
        svc.log.highlight(f'{self.target}: check target...')
        self._common_check()

        errs = 0
        self._shared.check()
        if errs > 0:
            svc.abort(f'{self.target} target_arduino: resolve errors')

    # --------------------
    ## generate Arduino target
    #
    # @return None
    def gen_target(self):
        self.tgt_build_dir = f'{svc.gbl.build_dir}/{self.target}-dir'

        svc.log.highlight(f'{self.target}: gen target, type:{self.target_type}')
        svc.log.line(f'   boardid  : {self._shared.boardid}')
        svc.log.line(f'   port     : {self._shared.avrdude_port}')
        svc.log.line(f'   baud_rate: {self._shared.avrdude_baudrate}')
        svc.log.line(f'   mcu      : {self._shared.mcu}')
        svc.log.line(f'   tgt dir  : {self.tgt_build_dir}')
        svc.log.line(f'   corelib  : {self._shared.corelib}')
        svc.log.line(f'   coredir  : {self._shared.coredir}')

        self._elffile = f'{svc.gbl.build_dir}/{self.target}.elf'
        self.add_clean(f'{self.target}.elf')

        self._eepfile = f'{svc.gbl.build_dir}/{self.target}.eep'
        self.add_clean(f'{self.target}.eep')

        self._hexfile = f'{svc.gbl.build_dir}/{self.target}.hex'
        self.add_clean(f'{self.target}.hex')

        self._gen_args()
        self._gen_init()
        self._gen_app()
        self._gen_link()
        self._gen_upload()

    # --------------------
    ## generate arguments for build
    def _gen_args(self):
        # create output build directory
        self._build_dirs[svc.gbl.build_dir] = 1

        for file in self.sources:  # pylint: disable=E1101
            _, _, dst_dir = self._get_obj_path(file)
            self._build_dirs[dst_dir] = 1

        self._writeln('')

    # --------------------
    ## generate initialization rule
    #
    # @return None
    def _gen_init(self):
        rule = f'{self.target}-init'
        self.add_rule(rule)

        self._gen_rule(rule, '', f'{self.target}: initialize for {svc.gbl.build_dir} build')

        for blddir in self._build_dirs:
            self._writeln(f'\t@mkdir -p {svc.osal.fix_path(blddir)}')
        self._writeln('')

    # --------------------
    ## generate build rule
    #
    # @return None
    def _gen_app(self):
        rule = f'{self.target}-build'
        self.add_rule(rule)

        # assumes _app_deps is only used in writeln()
        self._app_deps = ''
        build_deps = ''
        for file in self.sources:  # pylint: disable=E1101
            obj, mmd_inc, dst_dir = self._get_obj_path(file)

            # gen clean paths
            clean_path = dst_dir.replace(f'{svc.gbl.build_dir}/', '')
            self.add_clean(f'{svc.osal.fix_path(clean_path)}/*.o')
            self.add_clean(f'{svc.osal.fix_path(clean_path)}/*.d')

            self._writeln(f'-include {svc.osal.fix_path(mmd_inc)}')
            self._writeln(f'{svc.osal.fix_path(obj)}: {svc.osal.fix_path(file)}')
            self._writeln(f'\t{self._shared.cpp} {self._shared.cpp_opts} '
                          f'{self._compile_opts} {self._inc_dirs} '
                          f'{svc.osal.fix_path(file)} -o {svc.osal.fix_path(obj)}')
            self._app_deps += f'{svc.osal.fix_path(obj)} '
            build_deps += f'{svc.osal.fix_path(file)} '

        self._writeln('')

        self._gen_rule(rule, build_deps, f'{self.target}: build sketch')
        self._writeln('')

    # --------------------
    ## generate link rule
    #
    # @return None
    def _gen_link(self):
        rule = f'{self.target}-link'
        self.add_rule(rule)

        self._gen_rule(rule, f'{self._shared.core_tgt} {self._hexfile}',
                       f'{self.target}: link sketch and core')
        self._writeln('')

        # creates the ELF
        self._writeln(f'{svc.osal.fix_path(self._elffile)}: {self._app_deps} {svc.osal.fix_path(self._shared.corelib)}')
        self._writeln(f'\t{self._shared.cc} -Os -Wl,--gc-sections -mmcu={self._shared.mcu} '
                      f'-o {svc.osal.fix_path(self._elffile)} '
                      f'{self._app_deps} '
                      f'{svc.osal.fix_path(self._shared.corelib)} -lm')
        self._writeln('')

        # creates the EEP
        self._writeln(f'{svc.osal.fix_path(self._eepfile)}: {svc.osal.fix_path(self._elffile)}')
        self._writeln(f'\t{self._shared.obj_copy} -O ihex -j .eeprom --set-section-flags=.eeprom=alloc,load '
                      f'--no-change-warnings --change-section-lma .eeprom=0 '
                      f'{svc.osal.fix_path(self._elffile)} {svc.osal.fix_path(self._eepfile)}')
        self._writeln('')

        # creates the HEX
        self._writeln(f'{svc.osal.fix_path(self._hexfile)}: '
                      f'{svc.osal.fix_path(self._elffile)} {svc.osal.fix_path(self._eepfile)}')
        self._writeln(f'\t{self._shared.obj_copy} -O ihex -R .eeprom '
                      f'{svc.osal.fix_path(self._elffile)} {svc.osal.fix_path(self._hexfile)}')
        self._writeln('')

    # --------------------
    ## generate upload rule
    #
    # @return None
    def _gen_upload(self):
        rule = f'{self.target}-upload'
        # don't add rule

        avrdude_conf = f'{self._shared.avrdude_dir}/avrdude.conf'
        avrdude_protocol = self._shared.avrdude_protocol
        avrdude_args = f'-C{svc.osal.fix_path(avrdude_conf)} -p{self._shared.mcu} -c{avrdude_protocol} ' \
                       f'-P{self._shared.avrdude_port} -b{self._shared.avrdude_baudrate}'

        self._gen_rule(rule, f'{self.target}-link', f'{self.target}: upload to arduino')
        self._writeln(f'\t{self._shared.avrdude} -v {avrdude_args} -D '
                      f'-Uflash:w:{svc.osal.fix_path(self._hexfile)}:i '
                      f'-Ueeprom:w:{svc.osal.fix_path(self._eepfile)}:i')
        self._writeln('')



import os
import sys


# --------------------
## holds global information
class Gbl:
    ## hold the name of the build directory; default is debug
    build_dir = 'debug'
    ## holds the current OS name
    os_name = 'unknown'

    ## only do this once in cpp/gtest
    genned_empty_rule = False

    ## set clean commands to be quiet (prefix with '@')
    quiet_clean = False
    ## generate the compile_commands.json file
    gen_compile_commands = True

    # --------------------
    ## constructor
    def __init__(self):
        ## holds the current OS name
        self.os_name = 'unknown'
        if os.path.isfile('/sys/firmware/devicetree/base/model'):
            self.os_name = 'rpi'
        elif sys.platform == 'win32':
            self.os_name = 'win'
        elif sys.platform == 'darwin':
            self.os_name = 'macos'
        elif sys.platform == 'linux':
            self.os_name = 'ubuntu'
        else:
            print(f'unrecognized OS: "{sys.platform}"')

    # --------------------
    ## override the default build directory name
    #
    # @param build_dir  the new build directory
    # @return None
    def set_build_dir(self, build_dir):
        ## see Gbl.build_dir
        self.build_dir = build_dir



# --------------------
## base class for all targets
class TargetBaseMin:  # pylint: disable=too-many-instance-attributes
    # --------------------
    ## constructor
    #
    # @param target_name  the target name
    def __init__(self, target_name):
        ## target name
        self._target = target_name

        ## info for the clean rule for this target
        self._clean = {}
        ## help for this target
        self._help = {}
        ## list of rules for this target
        self._rules = []
        ## list of lines in the makefile for all aspects of this target
        self._lines = []

    # --------------------
    ## return the name of this target
    # @return the name of this target
    @property
    def target(self):
        return self._target

    # === target rules

    # --------------------
    ## add a new rule for this target
    #
    # @param rule  the name of the rule
    def add_rule(self, rule):
        self._rules.append(rule)

    # --------------------
    ## return the list of rules for this target
    # @return list of rules
    @property
    def rules(self):
        return self._rules

    # === clean rules

    # --------------------
    ## add clean target to list of patterns to clean
    #
    # @param pattern   the pattern to add
    # @return None
    def add_clean(self, pattern):
        if pattern not in self._clean:
            self._clean[pattern] = 1

    # --------------------
    ## return list of clean patterns for this target
    # @return return list of clean patterns
    @property
    def clean(self):
        return self._clean

    # === help text

    # --------------------
    ## add halp line for the given rule
    #
    # @param rule   the rule this help applies to
    # @param desc   the description for this help
    # @return None
    def _add_help(self, rule, desc):
        # note: replaces the help text if it already exists
        self._help[rule] = desc

    # --------------------
    ## return current help lines
    # @return list of help lines
    @property
    def help(self):
        return self._help

    # === gen functions

    # --------------------
    ## generate a rule
    #
    # @param rule   the rule's name
    # @param deps   the dependencies on this rule
    # @param desc   the description for this rule (comment in the makefile)
    def _gen_rule(self, rule, deps, desc):
        self._writeln(f'#-- {desc}')
        self._add_help(rule, desc)
        if deps:
            self._writeln(f'{rule}: {deps}')
        else:
            self._writeln(f'{rule}:')

    # --------------------
    ## generate lines to clean and generated directories and files given
    #
    # @return None
    def gen_clean(self):
        pass

    # === for writing to Makefile

    # --------------------
    ## return the list of lines for this target
    #
    # @return the list of lines
    @property
    def lines(self):
        return self._lines

    # --------------------
    ## save the given line to be generated later
    #
    # @param line  the line to write
    # @return None
    def _writeln(self, line):
        self._lines.append(line)



from falcon_logger import FalconLogger

from .arduino_shared import ArduinoShared
from .compile_cmd_json import CompileCmdJson
from .constants_version import ConstantsVersion
from .gbl import Gbl
from .osal import Osal
from .package_cpip import PackageCpip
from .package_opengl import PackageOpengl
from .svc import svc


# --------------------
## class to generate a makefile using a set of rules defined in python
class AlaMake:

    # --------------------
    ## constructor
    def __init__(self):
        svc.log = FalconLogger()
        svc.log.set_format('prefix')
        svc.gbl = Gbl()
        svc.osal = Osal  # not an instance, it is the class

        svc.log.line(f'OS name: {svc.gbl.os_name}')

        ## list of targets
        self._targets = []

        ## the file pointer to the makefile
        self._fp = None
        ## the current target
        self._tgt = None
        ## the list of rules
        self._rules = {}
        ## the list of help info
        self._help = {}
        svc.log.highlight('creating targets...')

    # --------------------
    ## return True if OS is RPI
    @property
    def is_rpi(self):
        return svc.gbl.os_name == 'rpi'

    # --------------------
    ## return True if OS is Windows/Msys2
    @property
    def is_win(self):
        return svc.gbl.os_name == 'win'

    # --------------------
    ## return True if OS is macoOS
    @property
    def is_macos(self):
        return svc.gbl.os_name == 'macos'

    # --------------------
    ## return True if OS is Ubuntu
    @property
    def is_ubuntu(self):
        return svc.gbl.os_name == 'ubuntu'

    # --------------------
    ## return the version string for pyalamake
    @property
    def version(self):
        return ConstantsVersion.version

    # --------------------
    ## return ref to svc global
    # @return svc.gbl
    @property
    def gbl(self):
        return svc.gbl

    # --------------------
    ## return ref to svc log
    # @return svc.log
    @property
    def log(self):
        return svc.log

    # --------------------
    ## return ref to svc OSAL
    # @return svc.osal
    @property
    def osal(self):
        return svc.osal

    # --------------------
    ## for UT only
    # @return ref to svc
    @property
    def ut_svc(self):
        return svc

    # === cfg related

    # --------------------
    ## generate quiet rm for clean command
    #
    # @param val (bool) set cfg value to True to run quiet, or False to be verbose
    # @return None
    def cfg_quiet_clean(self, val=True):
        svc.gbl.quiet_clean = val

    # --------------------
    ## generate compile_commands.json file
    #
    # @param val (bool) set cfg value to True to gen, or False to skip
    # @return None
    def cfg_gen_compile_commands(self, val=True):
        svc.gbl.gen_compile_commands = val

    # --------------------
    ## crate an arduino shared target
    #
    # @return reference to arduino shared object
    def create_arduino_shared(self):
        return ArduinoShared()

    # --------------------
    ## create a target with the given name and type.
    # The current recognized target types:
    #  * cpp - a C++ app or library
    #  * c   - a C app or library
    #  * gtest - a unit test app for test C++
    #  * arduino - an arduino app
    #  * arduino-core - an arduino core
    #  * swig - a swig translation to python/ruby
    #  * manual - generate manual compilation steps
    #
    # @param target_name   the name of the new target
    # @param target_type   the type of the target
    # @param shared        whether this is a shared target (e.g. Arduino Core)
    def create(self, target_name, target_type, shared=None):
        type_mapper = {
            # do not put arduino_core here
            'cpp': self._create_cpp,
            'cpp-lib': self._create_cpp_lib,
            'c': self._create_c,
            'c-lib': self._create_c_lib,
            'gtest': self._create_gtest,
            'arduino': self._create_arduino,
            'swig': self._create_swig,
            'manual': self._create_manual,
        }

        for tgt in self._targets:
            if target_name == tgt.target:
                svc.abort(f'target name is already in use: {target_name}')
                return None  # pragma: no cover

        if target_type == 'arduino-core':
            create_fn = self._create_arduino_core
        else:
            create_fn = type_mapper.get(target_type, None)

        if create_fn is None:
            svc.log.err(f'unknown target type: {target_type}')
            svc.abort(f'valid target types: {" ".join(type_mapper.keys())} ')
            return None  # pragma: no cover

        return create_fn(target_name, shared)

    # --------------------
    ## create a target with cpp type
    #
    # @param target_name   the name of the new target
    # @param shared        (unused) whether this is a shared target
    def _create_cpp(self, target_name, shared=None):  # pylint: disable=unused-argument
        from .target_cpp import TargetCpp
        svc.log.line(f'create: {target_name}')
        return TargetCpp.create(self._targets, target_name)

    # --------------------
    ## create a target with cpp lib type
    #
    # @param target_name   the name of the new target
    # @param shared        (unused) whether this is a shared target
    def _create_cpp_lib(self, target_name, shared=None):  # pylint: disable=unused-argument
        from .target_cpp_lib import TargetCppLib
        svc.log.line(f'create: {target_name}')
        return TargetCppLib.create(self._targets, target_name)

    # --------------------
    ## create a target with c type
    #
    # @param target_name   the name of the new target
    # @param shared        (unused) whether this is a shared target
    def _create_c(self, target_name, shared=None):  # pylint: disable=unused-argument
        from .target_c import TargetC
        svc.log.line(f'create: {target_name}')
        return TargetC.create(self._targets, target_name)

    # --------------------
    ## create a target with c lib type
    #
    # @param target_name   the name of the new target
    # @param shared        (unused) whether this is a shared target
    def _create_c_lib(self, target_name, shared=None):  # pylint: disable=unused-argument
        from .target_c_lib import TargetCLib
        svc.log.line(f'create: {target_name}')
        return TargetCLib.create(self._targets, target_name)

    # --------------------
    ## create a target with gtest type
    #
    # @param target_name   the name of the new target
    # @param shared        (unused) whether this is a shared target
    def _create_gtest(self, target_name, shared=None):  # pylint: disable=unused-argument
        from .target_gtest import TargetGtest
        svc.log.line(f'create: {target_name}')
        return TargetGtest.create(self._targets, target_name)

    # --------------------
    ## create a target with arduino type
    #
    # @param target_name   the name of the new target
    # @param shared        whether this is a shared target (e.g. Arduino Core)
    def _create_arduino(self, target_name, shared=None):
        svc.log.line(f'create: {target_name}')
        from .target_arduino import TargetArduino
        return TargetArduino.create(self._targets, target_name, shared=shared)

    # --------------------
    ## create a target with arduino core type
    #
    # @param target_name   the name of the new target
    # @param shared        whether this is a shared target (e.g. Arduino Core)
    def _create_arduino_core(self, target_name, shared=None):
        svc.log.line(f'create: {target_name}')
        from .target_arduino_core import TargetArduinoCore
        return TargetArduinoCore.create(self._targets, target_name, shared=shared)

    # --------------------
    ## create a target with swig type
    #
    # @param target_name   the name of the new target
    # @param shared        (unused) whether this is a shared target
    def _create_swig(self, target_name, shared=None):  # pylint: disable=unused-argument
        from .target_swig import TargetSwig
        svc.log.line(f'create: {target_name}')
        return TargetSwig.create(self._targets, target_name)

    # --------------------
    ## create a target with manual type
    #
    # @param target_name   the name of the new target
    # @param shared        (unused) whether this is a shared target
    def _create_manual(self, target_name, shared=None):  # pylint: disable=unused-argument
        from .target_manual import TargetManual
        svc.log.line(f'create: {target_name}')
        return TargetManual.create(self._targets, target_name)

    # --------------------
    ## find a package to add to this target.
    # Current packages recognized:
    #  cpip.* - see CPIP for available packages
    #  opengl - OpenGL package for graphics
    #
    # @param pkgname  the package name to search for
    # @return package info
    def find_package(self, pkgname):
        if pkgname.startswith('cpip.'):
            pkg = PackageCpip()
        elif pkgname == 'opengl':
            pkg = PackageOpengl()
        else:
            svc.log.err(f'unknown package: {pkgname}')
            svc.abort()
            return 'unknown'  # not needed, but stops IDE and pylint warnings

        return pkg.find(pkgname)

    # -------------------
    ## get the cross-platform port (e.g. COM3) based on the vid-pid of the USB port.
    #
    # @param vid_pid   the USB VID/PID value
    # @return the port with the VID/PID value in it, or None if not found
    def get_port(self, vid_pid):
        try:
            import serial
        except ModuleNotFoundError:
            svc.log.warn(f'{"get_port": <15}: module pyserial is not installed')
            return None

        import serial.tools.list_ports
        ports = serial.tools.list_ports.comports()
        for port, desc, hwid in sorted(ports):
            if vid_pid in hwid:
                svc.log.ok(f'{"get_port": <15}: found port {port}: {desc} [{hwid}]')
                return port
            # svc.log.dbg(f'port {port}: {desc} [{hwid}]')
        svc.log.err(f'{"get_port": <15}: vid-pid not found: "{vid_pid}"')
        svc.log.err(f'{"get_port": <15}: check USB is connected and powered on')
        return None

    # === makefile related

    # --------------------
    ## generate makefile
    #
    # @param path  the path to the makefile to generate; default: Makefile
    # @return None
    def makefile(self, ut_path=None):
        svc.log.highlight('generating targets...')
        self._gather_targets()

        self._gen_root_makefile()
        if ut_path:
            path = ut_path
        else:
            path = f'Makefile.{svc.gbl.os_name}'
        with open(path, 'w', encoding='utf-8') as self._fp:
            svc.log.highlight(f'generating makefile {path}...')
            self._gen_rules()
            self._gen_targets()
            self._gen_clean()
            self._gen_help()
        svc.log.line('done')

    # --------------------
    ## gather all targets
    #
    # @return None
    def _gather_targets(self):
        self._rules = {}
        for tgt in self._targets:
            # uncomment to debug
            # svc.log.dbg(f'   source   : {tgt.sources}')

            tgt.check()
            tgt.gen_target()
            tgt.gen_clean()
            self._rules[tgt.target] = tgt.rules

    # --------------------
    ## generate all rules
    #
    # @return None
    def _gen_rules(self):
        # gen rule for all
        rule = 'all'
        rules_str = ''
        for tgt, rules in self._rules.items():
            rules_str += f' {tgt} '
            rules_str += ' '.join(rules)
        self._writeln(f'.PHONY : all clean help {rules_str}')

        # has to be first target found otherwise clion can't parse it
        self._gen_rule(rule, rules_str, f'build {rule}')

        # generate a single rule to build each target in total
        for rule, rules_deps in self._rules.items():
            rules_str = ' '.join(rules_deps)
            self._gen_rule(rule, rules_str, f'build {rule}')

        self._writeln('')

    # --------------------
    ## generate all targets
    #
    # @return None
    def _gen_targets(self):
        ccj = CompileCmdJson()
        for tgt in self._targets:
            ccj.init()
            self._writeln(f'# ==== {tgt.target}')
            for line in tgt.lines:
                ccj.get_info(line)
                self._writeln(line)

        ccj.save()

    # --------------------
    ## generate help info
    #
    # @return None
    def _gen_help(self):
        bslash = '\\'
        self._add_help('help', 'this help info')

        # gather all the help
        all_help = {}
        all_help.update(self._help)
        for tgt in self._targets:
            all_help.update(tgt.help)

        self._writeln('help:')
        self._writeln(f'\t@printf "Available targets:{bslash}n"')
        last_rule = 'help'
        for rule, desc in sorted(all_help.items()):
            if rule.startswith(f'{last_rule}-'):
                rule_pfx = '  '
            else:
                last_rule = rule
                rule_pfx = ''
            desc2 = desc.replace('"', f'{bslash}"')
            self._writeln(f'\t@printf "  {rule_pfx}\x1b[32;01m{rule: <35}\x1b[0m {desc2}{bslash}n"')
        self._writeln(f'\t@printf "{bslash}n"')

    # --------------------
    ## generate lines to clean all generated files
    #
    # @return None
    def _gen_clean(self):
        rule = 'clean'
        clean_tgts = ''
        for tgt in self._targets:
            clean_tgts += f'{tgt.target}-clean '

        self._gen_rule(rule, clean_tgts, 'clean files')
        self._writeln('')

    # --------------------
    ## write a line to the makefile
    #
    # @param line  the line to write
    # @return None
    def _writeln(self, line):
        self._fp.write(line + '\n')

    # --------------------
    ## add a line for target help
    #
    # @param target   the target
    # @param desc     the help line
    # @return None
    def _add_help(self, target, desc):
        # TODO duplicate of function in target_base
        if target in self._help:
            svc.log.warn(f'add_help: target "{target}" already has description')
            svc.log.warn(f'   prev: {self._help[target]}')
            svc.log.warn(f'   curr: {desc}')
            svc.log.warn('   replacing...')
        self._help[target] = desc

    # --------------------
    ## generate a rule
    #
    # @param rule   the rule to generate
    # @param deps   the dependencies for this rule
    # @param desc   the help line
    # @return None
    def _gen_rule(self, rule, deps, desc):
        self._writeln(f'#-- {desc}')
        self._add_help(rule, desc)
        if deps:
            self._writeln(f'{rule}: {deps}')
        else:
            self._writeln(f'{rule}:')

    # --------------------
    ## generate root Makefile. This includes the correct file for the current OS
    #
    # @return None
    def _gen_root_makefile(self):
        content = ''
        content += '.phony: current_os\n'
        content += 'OS_NAME := $(shell uname -s)\n'
        content += 'ifeq ($(OS_NAME),Linux)\n'
        content += '    OS := ubuntu\n'
        content += '    include Makefile.ubuntu\n'
        content += 'else ifeq ($(OS_NAME),Darwin)\n'
        content += '    OS := macos\n'
        content += '    include Makefile.macos\n'
        content += 'else ifeq ($(OS_NAME),MSYS_NT-10.0-2610)\n'
        content += '    OS := msys2\n'
        content += '    include Makefile.msys2\n'
        content += 'else\n'
        content += '    # Handle other operating systems or raise an error\n'
        content += '    $(error Unsupported operating system: $(OS_NAME))\n'
        content += '    $(error select one of: Linux, Darwin, MSYS_NT-10.0-2610)\n'
        content += 'endif\n'
        content += '\n'
        content += 'current_os:\n'
        content += '\t@echo "Current OS: $(OS), uname: $(OS_NAME)"\n'
        path = 'Makefile'
        svc.log.highlight('generating root Makefile...')
        with open(path, 'w', encoding='utf-8') as self._fp:
            self._fp.write(content)



import os

from .list_param import ListParam
from .svc import svc
from .target_base_min import TargetBaseMin


# --------------------
## base class for all targets
class TargetBase(TargetBaseMin):  # pylint: disable=too-many-instance-attributes
    # --------------------
    ## constructor
    #
    # @param target_name  the target name
    def __init__(self, target_name):
        super().__init__(target_name)

        # note:  the lambda functions help Pycharm and pylint to resolve to dynamically allocated functions

        # === compile related

        ## list of sources property
        self.sources = []
        ## stub for add_sources function
        self.add_sources = lambda x: x
        ## list of source files parameter
        self._src_files_param = ListParam('sources', 'path',
                                          ## see function _sources_skip_if()
                                          self, self._sources_skip_if, None)

        ## list of include property
        self.include_directories = []
        ## stub for add_include_directories function
        self.add_include_directories = lambda x: x
        ## param string for include directories
        self._inc_dirs = ''
        ## list of include directories parameter
        self._inc_dirs_param = ListParam('include_directories', 'path',
                                         ## see function _update_inc_dirs
                                         self, None, self._update_inc_dirs)

        ## stub for compile_options property
        self.compile_options = []
        ## stub for add_compile_options function
        self.add_compile_options = lambda x: x
        ## param string for compilation options
        self._compile_opts = ''
        ## list of compile options parameter
        self._compile_opts_param = ListParam('compile_options', 'string',
                                             ## see function _update_compile_opts
                                             self, None, self._update_compile_opts)

        # === link related

        ## stub for link_options property
        self.link_options = []
        ## stub for add_link_options function
        self.add_link_options = lambda x: x
        ## list of link options
        self._link_opts = ''
        ## list of link options parameter
        self._link_opts_param = ListParam('link_options', 'string',
                                          ## see function _update_link_opts
                                          self, None, self._update_link_opts)

        ## stub for add_link_libraries property
        self.link_libraries = []
        ## stub for add_link_libraries function
        self.add_link_libraries = lambda x: x
        ## stub for link_files property
        self.link_files = []
        ## stub for add_link_files function
        self.add_link_files = lambda x: x
        ## param string for link libraries and files
        self._libs = ''
        ## list link libraries for this target; holds shortened library names i.e. no "lib" prefix, no ".a" extension
        self._link_libs_param = ListParam('link_libraries', 'string',
                                          ## see function _update_link_libs
                                          self, None, self._update_link_libs)
        ## list of paths to link files for this target; holds full path and library name
        self._link_files_param = ListParam('link_files', 'path',
                                           ## see function _update_link_libs
                                           self, None, self._update_link_libs)

        ## stub for add_link_directories property
        self.link_directories = []
        ## stub for add_link_directories function
        self.add_link_directories = lambda x: x
        ## param string for link paths
        self._link_paths = ''
        ## list link directories this target;  holds paths to search for libraries
        self._link_paths_param = ListParam('link_directories', 'path',
                                           ## see function _update_link_libs
                                           self, None, self._update_link_paths)

        ## info for cleaning the coverage generated
        self._clean_cov = {}

    # === help text

    # --------------------
    ## add halp line for the given rule
    # Note: overrides TargetBaseMin._add_help()
    #
    # @param rule   the rule this help applies to
    # @param desc   the description for this help
    # @return None
    def _add_help(self, rule, desc):
        if rule in self._help:
            svc.log.warn(f'add_help: target "{rule}" already has description')
            svc.log.warn(f'   prev: {self._help[rule]}')
            svc.log.warn(f'   curr: {desc}')
            svc.log.warn('   replacing...')
        self._help[rule] = desc

    # === source files

    # --------------------
    ## skip if a passed in source file value is an include file
    #
    # @param val   the source file to check
    # @return True if it is an include file (ends with .h) or False otherwise
    def _sources_skip_if(self, val):
        return val.endswith('.h')

    # === include directories

    # --------------------
    ## update include directories parameter to use in command line
    #
    # @return None
    def _update_inc_dirs(self):
        self._inc_dirs = ''
        for inc_dir in self._inc_dirs_param.values:
            self._inc_dirs += f'"-I{svc.osal.fix_path(inc_dir)}" '

    # === compilation/link flags

    # --------------------
    ## update the compile options parameter based on the current list of options
    #
    # @return None
    def _update_compile_opts(self):
        self._compile_opts = ''
        for opt in self._compile_opts_param.values:
            self._compile_opts += f'{opt} '

    # --------------------
    ## update the link options parameter based on the current list of options
    #
    # @return None
    def _update_link_opts(self):
        self._link_opts = ''
        for opt in self._link_opts_param.values:
            self._link_opts += f'{opt} '

    # === link libraries

    # --------------------
    ## update the link libraries command line info
    #
    # @return None
    def _update_link_libs(self):
        ## see base class for self._libs
        self._libs = ''

        ## see base class for self._link_libs
        for lib in self._link_libs_param.values:
            self._libs += f'-l{lib} '

        ## see base class for self._link_files
        for file in self._link_files_param.values:
            self._libs += f'"{file}" '

    # --------------------
    ## update the link directory paths command line info
    #
    # @return None
    def _update_link_paths(self):
        self._link_paths = ''
        for link_dir in self._link_paths_param.values:
            self._link_paths += f'-L{link_dir} '

    # === macos specific

    # --------------------
    ## update homebrew library and include directories for macOS.
    # ignored if not macOS.
    #
    # @return None
    def add_homebrew(self):
        if svc.gbl.os_name == 'macos':
            self.add_link_directories(svc.osal.homebrew_link_dirs())  # pylint: disable=E1101
            self.add_include_directories(svc.osal.homebrew_inc_dirs())  # pylint: disable=E1101

    # === gen functions

    # --------------------
    ## generate a path to an object file in this target
    #
    # @param file  the filename to use for the object file
    # @return obj: path to object file, dst_dir: path to the directory the object file is in
    def _get_obj_path(self, file):
        obj = f'{svc.gbl.build_dir}/{self.target}-dir/{file}.o'
        obj = obj.replace('//', '/')

        mmd_inc = f'{svc.gbl.build_dir}/{self.target}-dir/{file}.d'
        mmd_inc = mmd_inc.replace('//', '/')

        dst_dir = os.path.dirname(obj)
        return obj, mmd_inc, dst_dir

    # --------------------
    ## gen a rule used in passing args cpp/gtest cmd lines
    #
    # @return None
    def _gen_empty_rule(self):
        if svc.gbl.genned_empty_rule:
            return

        svc.gbl.genned_empty_rule = True
        self._writeln('# used to pass args in cpp and gtest cmd lines')
        self._writeln('%:')
        self._writeln('\t@:')

    # --------------------
    ## generate line to reset coverage info
    #
    # @param reset_rule  the name of the reset rule
    # @return None
    def _gen_reset_coverage(self, reset_rule):
        self._gen_rule(reset_rule, '', f'{self.target}: reset coverage info')

        for pattern in self._clean_cov:
            self._writeln(f'\trm -f {svc.gbl.build_dir}/{pattern}')
        self._writeln('')

    # --------------------
    ## generate lines to clean and generated directories and files given
    # Note: overrides TargetBaseMin.gen_clean
    #
    # @return None
    def gen_clean(self):
        clean_cov_rule = ''
        if self._clean_cov:
            reset_rule = f'{self.target}-cov-reset'
            self._gen_reset_coverage(reset_rule)
            clean_cov_rule = reset_rule

        rule = f'{self.target}-clean'
        self._gen_rule(rule, clean_cov_rule, f'{self.target}: clean files in this target')

        patterns = {}
        ## see clean property
        for pattern in self.clean:
            patterns[pattern] = 1

        quiet_pfx = ''
        if svc.gbl.quiet_clean:
            quiet_pfx = '@'
        for pattern in patterns:
            self._writeln(f'\t{quiet_pfx}rm -f {svc.gbl.build_dir}/{pattern}')
        self._writeln('')

    # --------------------
    ## various common checks for valid info
    #
    # @return None
    def _common_check(self):
        for file in self._src_files_param.values:
            if not os.path.isfile(file):
                svc.log.warn(f'{self.target}: source file {file} not found')

        for inc_dir in self._inc_dirs_param.values:
            if not os.path.isdir(inc_dir):
                svc.log.warn(f'{self.target}: include directory {inc_dir} not found')

        # _link_libs_param  # can't do, these may be generated
        # _link_files_param # can't do, these may be generated



from .svc import svc
from .target_base import TargetBase


# --------------------
## generate a C/C++ base functions
class TargetCCppBase(TargetBase):
    # create() is c/c++ specific

    # --------------------
    ## constructor
    #
    # @param target_name  the name of this target
    def __init__(self, target_name):
        super().__init__(target_name)

        ## the C/C++ target type
        self._target_type = 'unset'
        ## list of object files
        self._objs = ''
        ## compiler to use
        self._cxx = svc.osal.cpp_compiler()
        ## list of compile options
        self.add_compile_options('-D_GNU_SOURCE')  # pylint: disable=E1101

        ## list of build directories
        self._build_dirs = {}

    # target_type() is c/c++ specific
    # compiler() setter/getter are c/c++ specific

    # --------------------
    ## check target for any issues
    #
    # @return None
    def check(self):
        svc.log.highlight(f'{self.target}: check target...')
        self._common_check()

    # --------------------
    ## gen C++ target
    #
    # @return None
    def gen_target(self):
        svc.log.highlight(f'{self.target}: gen target, type:{self._target_type}')

        self._gen_args()
        self._gen_init()
        self._gen_app()
        self._gen_link()
        self._gen_run()

    # --------------------
    ## create output directory
    #
    # @return None
    def _gen_args(self):
        # create output build directory
        self._build_dirs[svc.gbl.build_dir] = 1

        for file in self.sources:  # pylint: disable=E1101
            _, _, dst_dir = self._get_obj_path(file)
            self._build_dirs[dst_dir] = 1

        self._writeln('')

    # --------------------
    ## gen initial content for C++ target
    #
    # @return None
    def _gen_init(self):
        rule = f'{self.target}-init'
        self.add_rule(rule)

        self._gen_rule(rule, '', f'{self.target}: initialize for {svc.gbl.build_dir} build')
        for bld_dir in self._build_dirs:
            self._writeln(f'\t@mkdir -p {svc.osal.fix_path(bld_dir)}')
        self._writeln('')

    # --------------------
    ## gen app build target
    #
    # @return None
    def _gen_app(self):
        rule = f'{self.target}-build'
        self.add_rule(rule)

        build_deps = ''
        for file in self.sources:  # pylint: disable=E1101
            obj, mmd_inc, dst_dir = self._get_obj_path(file)

            # gen clean paths
            clean_path = svc.osal.fix_path(dst_dir.replace(f'{svc.gbl.build_dir}/', ''))
            self.add_clean(f'{clean_path}/*.o')
            self.add_clean(f'{clean_path}/*.d')

            fixed_obj = svc.osal.fix_path(obj)
            fixed_file = svc.osal.fix_path(file)
            self._writeln(f'-include {svc.osal.fix_path(mmd_inc)}')
            self._writeln(f'{fixed_obj}: {fixed_file}')
            self._writeln(f'\t{self._cxx} -MMD -c {self._inc_dirs} {self._compile_opts} '
                          f'{fixed_file} -o {fixed_obj}')
            self._objs += f'{fixed_obj} '
            build_deps += f'{fixed_file} '

        self._writeln('')

        self._gen_rule(rule, self._objs, f'{self.target}: build source files')
        self._writeln('')

    # --------------------
    ## gen link target
    #
    # @return None
    def _gen_link(self):
        rule = f'{self.target}-link'
        self.add_rule(rule)

        exe = f'{svc.gbl.build_dir}/{self.target}'
        fixed_exe = svc.osal.fix_path(exe)
        self._writeln(f'{fixed_exe}: {self._objs}')
        # FYI "-Xlinker -Map=debug/x.map" to see the contents
        self._writeln(f'\t{self._cxx} {self._objs} {self._link_opts} {self._link_paths} {self._libs} '
                      f'-o {fixed_exe}')
        self._writeln('')
        ## see baseclass for definition of self.target
        self.add_clean(self.target)

        self._gen_rule(rule, exe, f'{self.target}: link')
        self._writeln('')

    # --------------------
    ## gen target to run c++ target
    #
    # @return None
    def _gen_run(self):
        rule = f'{self.target}-run'
        # don't add rule

        exe = f'{svc.gbl.build_dir}/{self.target}'

        self._gen_empty_rule()
        self._gen_rule(rule, f'{self.target}-link', f'{self.target}: run executable')
        self._writeln(f'\t@{svc.osal.fix_path(exe)} $(filter-out $@,$(MAKECMDGOALS))')
        self._writeln('')



import glob

from .svc import svc
from .target_base import TargetBase


# --------------------
## target for an Arduino Core
class TargetArduinoCore(TargetBase):
    # --------------------
    ## create an Arduino Core target
    #
    # @param targets      current list of targets
    # @param target_name  name of the new target
    # @param shared       the shared core info
    @classmethod
    def create(cls, targets, target_name, shared=None):
        impl = TargetArduinoCore(target_name, shared)
        targets.append(impl)
        return impl

    # --------------------
    ## constructor
    #
    # @param target_name  name of the new target
    # @param shared       the shared core info
    def __init__(self, target_name, shared):
        super().__init__(target_name)

        ## shared info
        self._shared = shared
        ## shared core target
        self._shared.core_tgt = target_name
        ## shared core directory
        self._shared.coredir = f'{svc.gbl.build_dir}/{self._shared.core_tgt}-dir'
        ## shared core library name
        self._shared.corelib_name = f'{self._shared.core_tgt}.a'
        ## shared core library path
        self._shared.corelib = f'{svc.gbl.build_dir}/{self._shared.corelib_name}'

        ## compile options to use
        self.add_compile_options(self._shared.debug_compile_opts)  # pylint: disable=E1101

        ## build directories to use
        self._build_dirs = {}

        ## library directory to use
        self._libdir = None
        ## list of object files to use
        self._objs = []
        ## list of include directories
        self._inc_dirs = []

    # --------------------
    ## the arduino core type
    # @return target type
    @property
    def target_type(self):
        return 'arduino_core'

    # --------------------
    ## check for various conditions in this target
    #
    # @return None
    def check(self):
        svc.log.highlight(f'{self.target}: check target...')
        self._common_check()

        errs = 0
        self._shared.check()
        if errs > 0:
            svc.abort(f'{self.target} target_arduino_core: resolve errors')

    # --------------------
    ## generate this target
    #
    # @return None
    def gen_target(self):
        svc.log.highlight(f'gen target {self.target}, type:{self.target_type}')

        self._gen_args()
        self._gen_init()

        self.add_clean(self._shared.corelib_name)

        self._libdir = svc.osal.arduino_core_libdir(self._shared.arduino_dir)

        self._inc_dirs = ''
        for incdir in self._shared.core_includes:
            self._inc_dirs += f' "-I{svc.osal.fix_path(incdir)}" '

        self._gen_core_compile()
        self._gen_core_link()
        self._writeln('')

    # --------------------
    ## generate argues for the build directory
    #
    # @return None
    def _gen_args(self):
        # create output build directory
        self._build_dirs[svc.gbl.build_dir] = 1
        self._build_dirs[self._shared.coredir] = 1

        self._writeln('')

    # --------------------
    ## generate init rule
    #
    # @return None
    def _gen_init(self):
        rule = f'{self.target}-init'
        self.add_rule(rule)

        self._gen_rule(rule, '', f'{self.target}: initialize for {svc.gbl.build_dir} build')
        for blddir in self._build_dirs:
            self._writeln(f'\t@mkdir -p {svc.osal.fix_path(blddir)}')
        self._writeln('')

    # --------------------
    ## generate core compilation rule
    #
    # @return None
    def _gen_core_compile(self):
        rule = f'{self.target}-build'
        self.add_rule(rule)

        cpp_files = self._gen_core_cpp_compile()
        c_files = self._gen_core_c_compile()

        # add clean rules; the clean function automatically adds the build_dir
        self.add_clean(f'{self._shared.core_tgt}-dir/*.o')
        self.add_clean(f'{self._shared.core_tgt}-dir/*.d')

        src_deps = ''
        for (src, _, _) in sorted(cpp_files + c_files):
            src_deps += f' {svc.osal.fix_path(src)}'

        self._gen_rule(rule, src_deps, f'{self.target}: compile arduino core source files')
        self._writeln('')

    # --------------------
    ## generate list of source files for C++ compilation
    #
    # @return list of source files
    def _gen_core_cpp_compile(self):
        src_files = []
        for src in glob.glob(f'{self._libdir}/*.cpp', recursive=True):
            obj = src.replace(self._libdir, self._shared.coredir) + '.o'
            mmd_inc = src.replace(self._libdir, self._shared.coredir) + '.d'
            src_files.append((src, mmd_inc, obj))

        for src, mmd_inc, obj in sorted(src_files):
            self._writeln(f'-include {svc.osal.fix_path(mmd_inc)}')
            self._writeln(f'{svc.osal.fix_path(obj)}: {svc.osal.fix_path(src)}')
            self._writeln(f'\t{self._shared.cpp} {self._shared.cpp_opts} '
                          f'{self._inc_dirs} {self._compile_opts} '
                          f'{svc.osal.fix_path(src)} -o {svc.osal.fix_path(obj)}')
            self._objs.append(obj)

        self._writeln('')

        return src_files

    # --------------------
    ## generate list of source files for C compilation
    #
    # @return list of source files
    def _gen_core_c_compile(self):
        src_files = []
        for src in glob.glob(f'{self._libdir}/*.c', recursive=True):
            obj = src.replace(self._libdir, self._shared.coredir) + '.o'
            mmd_inc = src.replace(self._libdir, self._shared.coredir) + '.d'
            src_files.append((src, mmd_inc, obj))

        for src, mmd_inc, obj in sorted(src_files):
            self._writeln(f'-include {svc.osal.fix_path(mmd_inc)}')
            self._writeln(f'{svc.osal.fix_path(obj)}: {svc.osal.fix_path(src)}')
            self._writeln(f'\t{self._shared.cc} {self._shared.cc_opts} '
                          f'{self._inc_dirs} {self._compile_opts} '
                          f'{svc.osal.fix_path(src)} -o {svc.osal.fix_path(obj)}')
            self._objs.append(obj)

        self._writeln('')

        return src_files

    # --------------------
    ## generate link rule
    #
    # @return None
    def _gen_core_link(self):
        rule = f'{self.target}-link'
        self.add_rule(rule)

        self._gen_rule(rule, self._shared.corelib, f'{self.target}: create arduino core library')
        self._writeln('')

        obj_deps = ''
        for obj in self._objs:
            obj_deps += f' {svc.osal.fix_path(obj)}'
        self._writeln(f'{self._shared.corelib}: {obj_deps}')
        self._writeln(f'\trm -f {self._shared.corelib}')
        for obj in sorted(self._objs):
            self._writeln(f'\t{self._shared.ar} rcs {self._shared.corelib} {svc.osal.fix_path(obj)}')



import json
import os
import re
import subprocess
from abc import ABC, abstractmethod

from .svc import svc


# --------------------
## base class to ensure both Null and real classes match up for function signatures
class CompileCmdJsonBase(ABC):
    # --------------------
    ## (skipped) initializer
    @abstractmethod
    def init(self):
        pass

    # --------------------
    ## (skipped) get info
    # @param line   (not used)
    @abstractmethod
    def get_info(self, line):
        pass

    # --------------------
    ## (skipped) save file
    @abstractmethod
    def save(self):
        pass


# --------------------
## return the correct class based on the configuration selected by the user
# use CamelCase to make it look like a class instantiation
# noinspection PyPep8Naming
def CompileCmdJson():  # pylint: disable=invalid-name
    if svc.gbl.gen_compile_commands:
        # generate the json file
        return _CompileCmdJson()

    # do not generate json file
    return _CompileCmdJsonNone()


# --------------------
## Null version of CompileCmdJson; used to skip generation of file
class _CompileCmdJsonNone(CompileCmdJsonBase):
    # --------------------
    ## (skipped) initializer
    def init(self):
        # TODO should this delete any existing compile_commands.json?
        pass

    # --------------------
    ## (skipped) get info
    # @param line   (not used)
    def get_info(self, line):
        pass

    # --------------------
    ## (skipped) save file
    def save(self):
        pass


# --------------------
## generate compile_commands.json file for clang-tidy linter
class _CompileCmdJson(CompileCmdJsonBase):
    # --------------------
    ## constructor
    def __init__(self):
        ## found a valid makefile line to parse or not
        self._found_it = False
        ## how many non-system compile objects found
        self._count = 0
        ## list of compile commands found
        self._compile_cmds = []
        ## info about the current command
        self._curr_cmd = {}

    # --------------------
    ## initialize
    def init(self):
        self._found_it = False

    # --------------------
    ## get information from the given makefile line
    #
    # @param line   the makefile line
    # @return None
    def get_info(self, line):
        if self._found_it:
            self._handle_line2(line)
            self._found_it = False
            self._curr_cmd = {}
            return

        m = re.search(r'(.*\.o):\s+(.*)', line)
        if not m:
            return

        obj_file = m.group(1)
        src_file = m.group(2)
        if src_file.startswith('/usr'):
            # exclude system includes/files
            return

        self._count += 1
        self._curr_cmd = {}
        self._compile_cmds.append(self._curr_cmd)
        self._handle_line1(obj_file, src_file)
        self._found_it = True

    # --------------------
    ## save the information to the compile_commands.json file
    #
    # @return None
    def save(self):
        # print(f'@@ DBG found: {self._count}')
        with open('compile_commands.json', 'w', encoding='utf-8') as fp:
            json.dump(self._compile_cmds, fp, indent=2, sort_keys=True)

    # --------------------
    ## get info from the first line of output
    #
    # @param obj_file  the obj file name
    # @param src_file  the src file name
    # @return None
    def _handle_line1(self, obj_file, src_file):
        # print(f'@@ DBG line1 {self._count: >2}: {obj_file} {src_file}')
        root_path = os.getcwd()
        self._curr_cmd['directory'] = root_path
        if src_file.startswith('/'):
            self._curr_cmd['file'] = src_file
        else:
            self._curr_cmd['file'] = os.path.join(root_path, src_file)

        if obj_file.startswith('/'):
            self._curr_cmd['output'] = obj_file
        else:
            self._curr_cmd['output'] = os.path.join(root_path, obj_file)

    # --------------------
    ## get info from the second line of output
    #
    # @param line  the second line
    # @return None
    def _handle_line2(self, line):
        tokens = []
        for token in line.strip().split():
            token = token.strip('"')
            if token in ['-MMD']:
                continue

            if not tokens:
                # first token is the command
                token = self._get_full_path(token)

            tokens.append(token)
        # print(f'@@ DBG line2 {self._count: >2}: {tokens}')
        self._curr_cmd['arguments'] = tokens

    # --------------------
    ## get the full path to the given command name
    #
    # @param cmd  the command to get info about
    # @return None
    def _get_full_path(self, cmd):
        result = subprocess.run(f'which {cmd}',
                                shell=True,  # Execute via shell
                                capture_output=True,  # Capture stdout and stderr
                                text=True,  # Decode output as string
                                check=True)  # Raise error for non-zero exit codes
        return result.stdout.strip()



import os
import pathlib

from .svc import svc
from .target_base import TargetBase


# --------------------
## generate a SWIG target
class TargetSwig(TargetBase):
    # --------------------
    ## create a SWIG target instance
    #
    # @param targets      current list of targets
    # @param target_name  name of new target to add
    @classmethod
    def create(cls, targets, target_name):
        impl = TargetSwig(target_name)
        targets.append(impl)
        return impl

    # --------------------
    ## constructor
    #
    # @param target_name  the name of this target
    def __init__(self, target_name):
        super().__init__(target_name)

        ## the swig target engine
        self._engine = None
        ## list of generated files
        self._genned = []
        ## list of additional swig options
        self._swig_opts = ''
        ## indicate if -c++ option
        self._cpp = False
        ## list of build directories
        self._build_dirs = {}

    # --------------------
    ## return target type
    #
    # @return cpp target
    @property
    def target_type(self):
        return 'swig'

    # --------------------
    ## set compiler to use
    # @return None
    def set_cpp(self):
        self._cpp = True

    # --------------------
    ## generate the given code for the swig target engine
    #
    # @param arg  the engine type
    # @return None
    def engine(self, arg):
        if arg == 'ruby':
            self._engine = '-ruby'
        elif arg == 'python':
            self._engine = '-python'
        elif arg == 'jsc':
            self._engine = '-javascript -jsc'
        elif arg == 'v8':
            self._engine = '-javascript -v8'
        elif arg == 'node':
            self._engine = '-javascript -node'
        elif arg == 'napi':
            self._engine = '-javascript -napi'

    # --------------------
    ## additional options for swig command
    #
    # @param opts  additional options
    # @return None
    def opts(self, opts):
        self._swig_opts = opts

    # --------------------
    ## check target for any issues
    #
    # @return None
    def check(self):
        svc.log.highlight(f'{self.target}: check target...')
        self._common_check()

    # --------------------
    ## gen C++ target
    #
    # @return None
    def gen_target(self):
        svc.log.highlight(f'{self.target}: gen target, type:{self.target_type}')

        self._gen_args()
        self._gen_init()
        self._gen_files()

    # --------------------
    ## create output directory
    #
    # @return None
    def _gen_args(self):
        # create output build directory
        self._build_dirs[svc.gbl.build_dir] = 1

        for file in self.sources:  # pylint: disable=E1101
            _, _, dst_dir = self._get_obj_path(file)
            self._build_dirs[dst_dir] = 1

        self._writeln('')

    # --------------------
    ## gen initial content for C++ target
    #
    # @return None
    def _gen_init(self):
        rule = f'{self.target}-init'
        self.add_rule(rule)

        self._gen_rule(rule, '', f'{self.target}: initialize for {svc.gbl.build_dir} build')
        for blddir in self._build_dirs:
            self._writeln(f'\t@mkdir -p {blddir}')
        self._writeln('')

    # --------------------
    ## gen source files based on swig
    #
    # @return None
    def _gen_files(self):
        rule = f'{self.target}-gen'
        self.add_rule(rule)

        if self._cpp:
            extension = 'cpp'
            self._swig_opts += '-c++ '
        else:
            extension = 'c'

        # should be list of ".i" files
        self._genned = []
        genned_files = ''
        for file in self.sources:  # pylint: disable=E1101
            gen_file = f'{svc.gbl.build_dir}/{self.target}-dir/{file}.{extension}'
            gen_file = gen_file.replace('//', '/')
            dst_dir = os.path.dirname(gen_file)
            pathlib.Path(dst_dir).mkdir(parents=True, exist_ok=True)

            # gen clean paths
            clean_path = gen_file.replace(f'{svc.gbl.build_dir}/', '')
            self.add_clean(f'{clean_path}/*.{extension}')

            self._writeln(f'{gen_file}: {file}')
            self._writeln(f'\tswig {self._engine} {self._swig_opts} -o {gen_file} {file} ')
            self._genned.append(gen_file)
            genned_files += f'{gen_file} '

        self._writeln('')

        self._gen_rule(rule, genned_files, f'{self.target}: genned source files')
        self._writeln('')
